#include "bakanet/core/ip_address.h"
#include "bakanet/core/ip_protocol.h"
#include "bakanet/core/socket.h"
#include "bakatools/container/types.h"
#include "bakatools/logging/assert.h"
#include "bakatools/logging/log.h"
#include <bakatools.h>
#include <bakanet.h>
#include <cerrno>
#include <cstdarg>
#include <cstddef>
#include <cstdio>
#include <functional>
#include <sys/epoll.h>
#include <thread>
#include <tuple>
#include <vector>

#include <string>

#define CREATE_MAPPING(var, ...) std::vector<std::pair<std::string, std::string>> var = { __VA_ARGS__ }
#define MAKE_RANGE(var, ...) range ( var, __VA_ARGS__)
#define BIND(listener, target) { listener, target }

#define MAX_EVENTS 10
#define MAX_CONNECTOR_WORKERS 8
#define MAX_LISTENER_CONNECTION 50
#define MAX_PACKET_SIZE 64000 

using namespace Bk;
using namespace Bk::Net;

void range(std::vector<std::pair<std::string, std::string>>& binds, std::string listener, std::string target, int min, int max) 
{ 
    for(int i = min; i <= max; i++) 
        binds.push_back({ listener + ":" + std::to_string(i), target + ":" + std::to_string(i) }); 
}

int setnonblocking(int sock)
{
    int result;
    int flags;

    flags = ::fcntl(sock, F_GETFL, 0);

    if (flags == -1)
    {
        return -1;  // error
    }

    flags |= O_NONBLOCK;

    result = fcntl(sock , F_SETFL , flags);
    return result;
}

struct Binder 
{
    Scope<Socket> src; 
    std::tuple<std::string, u16> listener;
    std::tuple<std::string, u16> target;

    Binder(Socket* src, std::tuple<std::string, u16> listener, std::tuple<std::string, u16> target)
    :src(src), listener(listener), target(target) { }
};


struct Bridge
{
    Scope<Socket> downstream, upstream; 

    Bridge(Socket* upstream, Socket* downstream)
    :upstream(upstream), downstream(downstream) { }
};

using PollCallback = std::function<void(int)>;

enum class PollEvent : i64
{
    None = 0,
    Accept = EPOLLIN | EPOLLET,
    Readable = EPOLLIN | EPOLLET,
    Writable = EPOLLOUT,
    Closed = EPOLLRDHUP,
    Error = EPOLLERR
};

class EventPoll 
{
    public:
        EventPoll(u16 maxEvents = 10)
        {
            instance = epoll_create1(0);
            BK_ASSERT(instance != -1);
            events.resize(maxEvents);
        }

        void add(int fd, int nbEvents = 0, ...)
        {
            va_list args;
            va_start(args, nbEvents);
            struct epoll_event ev;
            ev.events = 0;
            for(int i = 0; i < nbEvents; i++) ev.events = ev.events | va_arg(args, i64);
            ev.data.fd = fd;
            BK_STATIC_MSG_ASSERT(epoll_ctl(instance, EPOLL_CTL_ADD, fd, &ev) != -1, "epoll_ctl: listen_sock");
        }

        void on(PollEvent event, PollCallback callback)
        {
            callbacks[event] = callback;
        }

        void wait(int timeout = -1) 
        {
            int waiters = epoll_wait(instance, events.data(), events.size(), timeout);
            BK_STATIC_MSG_ASSERT(waiters != -1, "epoll_ctl: listen_sock");
            for(int i = 0; i < waiters; i++)
            {
                if((events[i].events & (i64)PollEvent::Closed) != 0 && callbacks[PollEvent::Closed]) callbacks[PollEvent::Closed](events[i].data.fd);
                else if((events[i].events & (i64)PollEvent::Readable) != 0 && callbacks[PollEvent::Readable]) callbacks[PollEvent::Readable](events[i].data.fd);
                else if((events[i].events & (i64)PollEvent::Accept) != 0 && callbacks[PollEvent::Accept]) callbacks[PollEvent::Accept](events[i].data.fd);
                else if((events[i].events & (i64)PollEvent::Writable) != 0 && callbacks[PollEvent::Writable]) callbacks[PollEvent::Writable](events[i].data.fd);
                else if((events[i].events & (i64)PollEvent::Error) != 0 && callbacks[PollEvent::Error]) callbacks[PollEvent::Error](events[i].data.fd);
            }
        }

    private:
        std::vector<struct epoll_event> events;
        int instance = -1;
        std::map<PollEvent, PollCallback> callbacks;

};

class Connector 
{
    public:
        Connector()
        {
            bridges.reserve(10);
            ePoll.on(PollEvent::Accept, [this](int fd)
            {
                int index = findBridge(fd);
                if(index > -1) 
                {
                    auto& bridge = bridges[index];
                    if (bridge.upstream->get_raw_socket() == fd) 
                    { 
                        std::vector<char> packet = bridge.upstream->obtain(MAX_PACKET_SIZE);

                        if(packet.size() > 0) bridge.downstream->emit(packet);
                        return;
                    }
                    std::vector<char> packet = bridge.downstream->obtain(MAX_PACKET_SIZE);
                    if(packet.size() > 0) bridge.upstream->emit(packet);
                }
            });

            ePoll.on(PollEvent::Closed, [this](int fd)
            {
                int i = findBridge(fd);
                BK_INFO("Deleted bridge N° {0} - Remaining bridges {1}\n", i, bridges.size() -1);
                if(i > -1) bridges.erase(bridges.begin() + i);
            });
        }

        ~Connector()
        {
            if(running) {
                stop();
                worker.join();
            }
        }

        int findBridge(int fd)
        {
            for(int i = 0; i < bridges.size(); i++)
            {
                if (bridges[i].upstream->get_raw_socket() == fd || bridges[i].downstream->get_raw_socket() == fd) {
                    return i;
                } 
            }
            return -1;
        }

        void start()
        {
            if(!running)
            {
                running = true;
                worker = std::thread([this]
                {
                    while(running)
                    {
                        std::unique_lock<std::mutex> lock(resource_mutex);
                        condition.wait(lock, [this] {
                            return !bridges.empty() || !running;
                        });

                        if(!running || bridges.empty()) return; 
                        
                        ePoll.wait();
                    }
                });
            }
        }

        void stop()
        {
            std::unique_lock<std::mutex> lock(resource_mutex);
            running = false;
            condition.notify_one();
        }

        void add(Bridge bridge)
        {
            {
                std::unique_lock<std::mutex> lock(resource_mutex);
                BK_STATIC_ASSERT(bridge.upstream->conn());
                setnonblocking(bridge.upstream->get_raw_socket());
                ePoll.add(bridge.upstream->get_raw_socket(), 2, PollEvent::Readable, PollEvent::Closed);
                ePoll.add(bridge.downstream->get_raw_socket(), 2, PollEvent::Readable, PollEvent::Closed);
                bridges.push_back(std::move(bridge));
            }
            BK_INFO("Added bridge - Total bridges {0}\n", bridges.size() - 1);
            condition.notify_one();
        }

    private:
        std::thread worker;

        std::condition_variable condition;
        std::mutex resource_mutex;

        bool running = false;
        int instance;
        EventPoll ePoll;
        std::vector<Bridge> bridges;
};

std::tuple<std::string, u32> parseConnection(std::string conn)
{
    size_t pos = conn.find_first_of(":");
    if(pos != std::string::npos) {
        auto ip = conn.substr(0, pos);
        auto port = std::stoi(conn.substr(pos + 1, conn.length()));
        return std::tuple(ip, port);
    } else if (conn.length()) return std::tuple("", std::stoi(conn.substr(0, conn.length())));
    return std::tuple("", 0);
}

class TcpProxy 
{
    public:
    TcpProxy(const std::vector<std::pair<std::string, std::string>>& proxyBinds)
    {
        binds.reserve(proxyBinds.size());
        BK_INFO("Starting binds (count: {0})", proxyBinds.size());
        int i = 0;
        for (auto& proxyBind : proxyBinds)
        {
            std::tuple<std::string, u16> listener = parseConnection(proxyBind.first);
            std::tuple<std::string, u16> target = parseConnection(proxyBind.second);
            BK_INFO("{0}. Buiding binder [{1}:{2}] -> [{3}:{4}]", 
                i + 1, 
                std::get<0>(listener), 
                std::get<1>(listener), 
                std::get<0>(target), 
                std::get<1>(target)
            );
            binds.insert(binds.begin() + i, Binder(
                    Socket::create(std::get<0>(listener), std::get<1>(listener), IpProtocol::TCP), 
                    listener, 
                    target
                )
            );
            BK_ASSERT(binds[i].src->init());
            BK_ASSERT(binds[i].src->start(MAX_LISTENER_CONNECTION));
            ePoll.add(binds[i].src->get_raw_socket(), 1, PollEvent::Accept);
            i++;
        }

        ePoll.on(PollEvent::Accept, [this] (int fd)
        {
            auto& bind = binds[findBinder(fd)];
            Socket* conn = bind.src->ack();
            Socket* target = Socket::create(IpAddress(std::get<0>(bind.target)), std::get<1>(bind.target), IpProtocol::TCP);
            BK_INFO("New connection");
            BK_INFO("Connection info :");
            BK_INFO("PORT BIND  [{0}:{1}]", std::get<1>(bind.listener), std::get<1>(bind.target));
            BK_INFO("TARGET     [{0}]", target->get_ip());
            BK_INFO("LISTENER   [{0}]", bind.src->get_ip());
            BK_INFO("CLIENT     [{0}]", conn->get_ip());
            setnonblocking(conn->get_raw_socket());
            if(connectorPtr >= connectors.size()) connectorPtr = 0;
            BK_INFO("Forwarding to connector N°{0}\n", connectorPtr);
            connectors[connectorPtr].add(Bridge(target, conn));
            connectorPtr++;
        });
    }

    int findBinder(int fd)
    {
        for(int i = 0; i < binds.size(); i++)
        {
            if (binds[i].src->get_raw_socket() == fd) {
                return i;
            } 
        }
        return -1;
    }

    void start()
    {
        running = true;
        for(auto& connector : connectors)
        {
            connectorPtr++;
            BK_INFO("Starting connector N°{0}", connectorPtr);
            connector.start();
        }
        printf("\n");
        while (running) {
            ePoll.wait();
        } 
    }

    private:
        bool running = false;
        std::vector<Binder> binds;
        EventPoll ePoll;
        std::array<Connector, MAX_CONNECTOR_WORKERS> connectors;
        int connectorPtr = 0;
};

int main()
{   
    Log::Init("Bakanet");
    BK_INFO("Epoll events types");
    BK_INFO("EPOLLIN    {0} {0:b}", (u32)EPOLLIN);
    BK_INFO("EPOLLOUT   {0} {0:b}", (u32)EPOLLOUT);
    BK_INFO("EPOLLET    {0} {0:b}", (u32)EPOLLET);
    BK_INFO("EPOLLHUP   {0} {0:b}", (u32)EPOLLHUP);
    BK_INFO("EPOLLERR   {0} {0:b}", (u32)EPOLLERR);
    BK_INFO("EPOLLRDHUP {0} {0:b}", (u32)EPOLLRDHUP);
    printf("\n");

    CREATE_MAPPING(proxyBinds, 
        BIND("192.168.1.128:25000", "192.168.1.128:25001"),
        BIND("192.168.1.128:25001", "192.168.1.128:25002"),
        BIND("192.168.1.128:25002", "192.168.1.128:25003"),
        BIND("192.168.1.128:25003", "192.168.1.128:25004"),
        BIND("192.168.1.128:25004", "192.168.1.128:25005"),
        BIND("192.168.1.128:25005", "192.168.1.128:25006"),
        BIND("192.168.1.128:25006", "192.168.1.128:25007"),
        BIND("192.168.1.128:25007", "192.168.1.159:22")
    );
    TcpProxy initiator(proxyBinds);
    initiator.start();
    return 0;
}